package com.mowerm.opentools.clippy;

import java.awt.*;
import java.awt.datatransfer.*;
import java.io.*;

import com.borland.primetime.properties.*;
import com.borland.primetime.vfs.*;

/**
 *  The <code>BufferManager</code> handles access to the clipboard buffers.<p>
 *
 *  This source code is hereby placed in the public domain.<p>
 *
 *
 *
 *@author     Matt Mower <matt_mower@hotmail.com>
 *@created    June 28, 2001
 *@version    1.0
 */

public class BufferManager {
	/**
	 *  Index of buffer which actions currently reference
	 */
	protected int currentBuffer = 0;

	/**
	 *  Actual clipboards the buffers represent
	 */
	protected Clipboard[] clipboards = null;

	/**
	 *  A dummy clipboard owner - all clipboards require one
	 */
	protected ClipboardOwner clipboardOwner = new ClipboardObserver();

	/**
	 *  A locking object
	 */
	private static Object systemClipboardLocker = new Object();


	/**
	 *  Materialize a new <code>BufferManager</code> .<p>
	 *
	 *
	 */
	public BufferManager() {
		loadBuffers();
	}



	/**
	 *  Makes the specified buffer the current buffer.<p>
	 *
	 *
	 *
	 *@param  _bufferIndex  New current buffer index
	 */
	public void setCurrentBuffer( int _bufferIndex ) {
		if( 0 <= _bufferIndex && _bufferIndex < getBufferCount() ) {
			currentBuffer = _bufferIndex;
		}
	}


	/**
	 *  Set the <code>Clipboard</code> representation for the specified buffer.<p>
	 *
	 *
	 *
	 *@param  _bufferIndex  Index of buffer
	 *@param  _buffer       New buffer representation
	 */
	public void setBuffer( int _bufferIndex, Clipboard _buffer ) {
		getBuffers()[_bufferIndex] = _buffer;
	}


	/**
	 *  Sets the content of the current buffer.<p>
	 *
	 *
	 *
	 *@param  _contents  New content of buffer
	 */
	public void setBufferContents( Transferable _contents ) {
		setBufferContents( getBufferIndex(), _contents );
	}


	/**
	 *  Sets the contents of the specified buffer.<p>
	 *
	 *
	 *
	 *@param  _bufferIndex  Index of buffer
	 *@param  _contents     New contents of buffer
	 */
	public void setBufferContents( int _bufferIndex, Transferable _contents ) {
		getBuffer( _bufferIndex ).setContents( _contents, clipboardOwner );
	}


	/**
	 *  Gets the <code>Clipboard</code> representing the specified buffer.<p>
	 *
	 *
	 *
	 *@param  _bufferIndex  Index of buffer
	 *@return               Content of buffer
	 */
	public Clipboard getBuffer( int _bufferIndex ) {
		return getBuffers()[_bufferIndex];
	}


	/**
	 *  Get the number of buffers.<p>
	 *
	 *
	 *
	 *@return    number of buffers
	 */
	public int getBufferCount() {
		return clipboards.length;
	}


	/**
	 *  Gets the <code>Clipboard</code> representing the current buffer.<p>
	 *
	 *
	 *
	 *@return    <code>Clipboard</code> buffer
	 */
	public Clipboard getBuffer() {
		return getBuffer( getBufferIndex() );
	}


	/**
	 *  Gets the index of the current buffer.<p>
	 *
	 *
	 *
	 *@return    Current buffer index
	 */
	public int getBufferIndex() {
		return currentBuffer;
	}


	/**
	 *  Gets the index of the current buffer as it should be displayed on-screen
	 *  (we chose 1-based indices for screen and 0-based for internal use).<p>
	 *
	 *
	 *
	 *@return    Current buffer index for on-screen use
	 */
	public int getBufferDisplayNumber() {
		return getBufferIndex() + 1;
	}


	/**
	 *  Gets the content of the current buffer as a <code>Transferable</code>
	 *  instance.<p>
	 *
	 *
	 *
	 *@return    Content of the current buffer.
	 */
	public Transferable getBufferContents() {
		return getBuffer().getContents( clipboardOwner );
	}


	/**
	 *  Gets the content of the specified buffer as a <code>Transferable</code>
	 *  instance.<p>
	 *
	 *
	 *
	 *@param  _bufferIndex  Index of buffer
	 *@return               Content of the specified buffer
	 */
	public Transferable getBufferContents( int _bufferIndex ) {
		return getBuffer( _bufferIndex ).getContents( clipboardOwner );
	}


	/**
	 *  Gets the <code>ClipboardOwner</code>.<p>
	 *
	 *
	 *
	 *@return    <code>ClipboardOwner</code>
	 */
	public ClipboardOwner getClipboardOwner() {
		return clipboardOwner;
	}


	/**
	 *  Gets a 60-character or less summary of the content of the current buffer.
	 *  <p>
	 *
	 *
	 *
	 *@return    Summary of the buffer content
	 */
	public String getBufferSummary() {
		final String _content = getBufferAsString();
		return _content.length() < 61 ? _content : _content.substring( 0, 60 );
	}


	/**
	 *  Gets the content of the current buffer in <code>String</code> form.<p>
	 *
	 *
	 *
	 *@return    <code>String</code> representation of current buffer content
	 */
	public String getBufferAsString() {
		return getBufferAsString( getBufferIndex() );
	}


	/**
	 *  Gets the specified buffer content in string form.<p>
	 *
	 *
	 *
	 *@param  _bufferIndex  Index of buffer
	 *@return               <code>String</code> representation of specified buffer
	 *      content
	 */
	public String getBufferAsString( int _bufferIndex ) {
		final Transferable _content = getBufferContents( _bufferIndex );
		if( _content != null ) {
			final DataFlavor[] _flavours = _content.getTransferDataFlavors();
			if( _flavours == null && _flavours.length < 1 ) {
				return "<gibberish>";
			}
			try {
				return _content.getTransferData( _flavours[0] ).toString();
			} catch( Exception _ex ) {
				return "<unintelligible>";
			}
		} else {
			return "<empty>";
		}
	}


	/**
	 *  Gets the content of the current buffer as a raw unformatted string.<p>
	 *
	 *
	 *
	 *@return    <String> containing buffer content
	 */
	public String getBufferAsRawString() {
		return getBufferAsRawString( getBufferIndex() );
	}


	/**
	 *  Gets the content of the specified buffer as a raw unformatted string.<p>
	 *
	 *
	 *
	 *@param  _bufferIndex  Index of buffer
	 *@return               <String> containing buffer content
	 */
	public String getBufferAsRawString( int _bufferIndex ) {
		final Transferable _content = getBufferContents( _bufferIndex );
		if( _content != null ) {
			final DataFlavor[] _flavours = _content.getTransferDataFlavors();
			if( _flavours == null && _flavours.length < 1 ) {
				return null;
			}
			try {
				return _content.getTransferData( _flavours[0] ).toString();
			} catch( Exception _ex ) {
				return null;
			}
		} else {
			return null;
		}
	}


	/**
	 *  Swaps the underlying representation of two buffers.<p>
	 *
	 *
	 *
	 *@param  _indexA  Index of first buffer
	 *@param  _indexB  Index of second buffer
	 */
	public void swapBuffers( int _indexA, int _indexB ) {
		Clipboard _temp = getBuffer( _indexA );
		setBuffer( _indexA, getBuffer( _indexB ) );
		setBuffer( _indexB, _temp );
	}


	/**
	 *  Treating the buffers as a ring, advances the buffers around the ring so
	 *  that the last buffer becomes the first and all other advance one place.<p>
	 *
	 *
	 */
	public void advanceBuffers() {
		Clipboard _droppedBuffer = getBuffer( getBufferCount() - 1 );
		for( int _bufferIndex = getBufferCount() - 1; _bufferIndex > 0; --_bufferIndex ) {
			setBuffer( _bufferIndex, getBuffer( _bufferIndex - 1 ) );
		}
		setBuffer( 0, _droppedBuffer );
	}


	/**
	 *  Copies the specified buffer to the system clipboard allowing its content to
	 *  be pasted by other applications.<p>
	 *
	 *
	 */
	public void copyBufferToSystemClipboard() {
		synchronized( systemClipboardLocker ) {
			Transferable _content = getBufferContents();
			if( _content == null ) {
				// Empty buffer
				_content = new StringSelection( "" );
			}
			Toolkit _kit = Toolkit.getDefaultToolkit();
			Clipboard _clipboard = _kit.getSystemClipboard();
			Clippy.debug( "Clipboard class = " + _clipboard.getClass().getName() );
			Clippy.debug( "Calling #setContents( " + _content.toString() + ", " + clipboardOwner.toString() + " )" );
			_clipboard.setContents( _content, clipboardOwner );
			Clippy.debug( "Clipboard contents set" );
		}
	}


	/**
	 *  Copies the content of the system clipboard into the current buffer.
	 *  Allowing content copied from other applications to appear in Clippy.<p>
	 *
	 *
	 */
	public void copySystemClipboardToBuffer() {
		copySystemClipboardToBuffer( getBufferIndex() );
	}


	/**
	 *  Copies the content of the system clipboard to the specified buffer.<p>
	 *
	 *
	 *
	 *@param  _bufferIndex  Index of buffer
	 */
	public void copySystemClipboardToBuffer( int _bufferIndex ) {
		synchronized( systemClipboardLocker ) {
			Transferable _content = Toolkit.getDefaultToolkit().getSystemClipboard().getContents( clipboardOwner );
			if( _content != null ) {
				setBufferContents( _bufferIndex, _content );
			}
		}
	}


	/**
	 *  Advances the current buffer index. Treats the buffers as a ring, hence
	 *  advancing beyond the last buffer will reset the pointer to the first
	 *  buffer.<p>
	 *
	 *
	 */
	public void selectNextBuffer() {
		++currentBuffer;
		if( currentBuffer == getBufferCount() ) {
			currentBuffer = 0;
		}
	}


	/**
	 *  Saves the content of the buffers for use in a subsequent session.<p>
	 *
	 *
	 */
	public void saveBuffers() {
		try {
			Url _buffersUrl = PropertyManager.getSettingsUrl( "clippy.buffers" );
			ObjectOutputStream _saveStream = new ObjectOutputStream( new BufferedOutputStream( VFS.getOutputStream( _buffersUrl ) ) );
			writeBuffers( _saveStream );
			_saveStream.flush();
			_saveStream.close();
		} catch( Exception _ex ) {
			_ex.printStackTrace();
		}
	}


	/**
	 *  Loads the content of the buffers, saved from a previous session.<p>
	 *
	 *  If the buffers cannot be loaded they are initialized to be empty.<p>
	 *
	 *
	 */
	public void loadBuffers() {
		Url _buffersUrl = PropertyManager.getSettingsUrl( "clippy.buffers" );

		Clippy.debug( "Clippy is attempting to load its buffers from last session [" + _buffersUrl.getFullName() + "]" );

		try {
			if( VFS.exists( _buffersUrl ) ) {
				ObjectInputStream _loadStream = new ObjectInputStream( new BufferedInputStream( VFS.getInputStream( _buffersUrl ) ) );
				readBuffers( _loadStream );
				_loadStream.close();
				// Keep this property in sync so we don't have accidents later!
				ClippyPropertiesGroup.BUFFER_COUNT.setInteger( clipboards.length );
			} else {
				Clippy.debug( "-->No buffers file, initializing empty buffers" );
				initBuffers();
			}
		} catch( Exception _ex ) {
			// If we cannot load then get rid of the broken buffers file & initialize the buffers
			// as normal
			try {
				VFS.delete( _buffersUrl );
			} catch( IOException _ioex ) {
				_ioex.printStackTrace();
			}
			initBuffers();
			_ex.printStackTrace();
		}
	}


	/**
	 *  Gets an array of <code>Clipboard</code> instances representing the
	 *  underlying storage for the buffers.<p>
	 *
	 *
	 *
	 *@return    Clipboards
	 */
	protected Clipboard[] getBuffers() {
		return clipboards;
	}


	/**
	 *  Writes the buffer content as a stream of serialized objects.<p>
	 *
	 *
	 *
	 *@param  _saveStream      <code>ObjectOutputStream</code> to write to
	 *@exception  IOException  if an I/O error occurs
	 */
	protected void writeBuffers( ObjectOutputStream _saveStream ) throws IOException {
		_saveStream.writeInt( clipboards.length );
		for( int _index = 0; _index < clipboards.length; ++_index ) {
			_saveStream.writeObject( getBufferAsString( _index ) );
		}
	}


	/**
	 *  Reads the buffer content from a stream of serialized objects.<p>
	 *
	 *
	 *
	 *@param  _loadStream                 <code>ObjectInputStream> to read from
	 *
	 *
	 *
	 *@exception  IOException             if an I/O error occurs
	 *@exception  ClassNotFoundException  if a Class cannot be found
	 */
	protected void readBuffers( ObjectInputStream _loadStream ) throws IOException, ClassNotFoundException {
		int _bufferCount = _loadStream.readInt();
		initBuffers( _bufferCount );
		for( int _index = 0; _index < _bufferCount; ++_index ) {
			String _content = ( String ) _loadStream.readObject();
			Clippy.debug( "Clippy : Clipboard " + ( _index + 1 ) + " = '" + _content + "'" );
			StringSelection _selection = new StringSelection( _content );
			clipboards[_index].setContents( _selection, clipboardOwner );
		}
		// Dont call Clippy.setBufferCount() as this will lead to a resize
		// and eventually a StackOverflowError!
		ClippyPropertiesGroup.BUFFER_COUNT.setInteger( _bufferCount );
	}


	/**
	 *  Initializes the buffers to an empty state.<p>
	 *
	 *
	 */
	protected void initBuffers() {
		initBuffers( Clippy.getBufferCount() );
	}


	/**
	 *  Initializes the specified number of buffers to an empty state.<p>
	 *
	 *
	 *
	 *@param  _bufferCount  Description of Parameter
	 */
	protected void initBuffers( int _bufferCount ) {
		clipboards = new Clipboard[_bufferCount];
		for( int _i = 0; _i < clipboards.length; ++_i ) {
			clipboards[_i] = new Clipboard( "Buffer<" + _i + ">" );
		}
	}


	/**
	 *  Misnamed method, doesn't resize the buffers themselves but changes the
	 *  number of buffers.<p>
	 *
	 *
	 *
	 *@param  _bufferCount  New number of buffers
	 */
	protected void resizeBuffers( int _bufferCount ) {
		Clipboard[] _newBuffers = new Clipboard[_bufferCount];

		int _nrToKeep = ( _bufferCount > getBufferCount() ) ? getBufferCount() : _bufferCount;
		for( int _index = 0; _index < _bufferCount; ++_index ) {
			if( _index < _nrToKeep ) {
				_newBuffers[_index] = clipboards[_index];
			} else {
				_newBuffers[_index] = new Clipboard( "Buffer<" + _index + ">" );
			}
		}

		// Re-assign the buffer list
		clipboards = _newBuffers;

		// Re-select the first buffer
		setCurrentBuffer( 0 );
	}


	/**
	 *  An implementation of <code>ClipboardOwner</code> provided only because all
	 *  Clipboards must have an owner and we create a <code>Clipboard</code> to
	 *  represent each buffer.<p>
	 *
	 *
	 *
	 *@author     Matt Mower <matt@metadyne.uk.com>
	 *@created    June 28, 2001
	 *@version    $Revision$
	 */
	protected static class ClipboardObserver implements ClipboardOwner {
		/**
		 *  We don't care about this.
		 *
		 *@param  clipboard  Clipboard
		 *@param  contents   Content
		 */
		public void lostOwnership( Clipboard clipboard, Transferable contents ) {
			Clippy.debug( "##lost ownership" );
		}
	}
}
